/**
 * Copyright (c) 2025 Herbert Poetzl
 *
 * ER-TFT028-4   4-wire SPI @62.5MHz
 *
 */

#include <stdio.h>
#include <stdint.h>

#include "pico/stdlib.h"
#include "pico/rand.h"
#include "hardware/pio.h"
#include "hardware/dma.h"

#include "spigen.pio.h"


const uint pin_d_scl = 2;
const uint pin_d_sda = 3;
const uint pin_d_sdo = 4;
const uint pin_d_rsn = 5;
const uint pin_d_csn = 6;
const uint pin_d_dcn = 7;


#define DISPLAY_WIDTH  240
#define DISPLAY_HEIGHT 320

#define H_OFFSET       0  // Horizontal offset (adjust if needed)
#define V_OFFSET       0  // Vertical offset (adjust if needed)

#define ILI9341_NOP     0x00
#define ILI9341_SWRESET 0x01
#define ILI9341_SLPOUT  0x11
#define ILI9341_DISPON  0x29
#define ILI9341_CASET   0x2A
#define ILI9341_PASET   0x2B
#define ILI9341_RAMWR   0x2C
#define ILI9341_MADCTL  0x36
#define ILI9341_COLMOD  0x3A


uint16_t hueToRGB565(uint16_t hue) {
    // Scale hue to 6 segments (0-1535)
    uint32_t scaledHue = ((uint32_t)hue * 6) >> 8;  // 65536*6/256=1536 steps
    uint8_t segment = scaledHue >> 8;  // 0-5 (6 segments)
    uint8_t x = scaledHue & 0xFF;      // Fractional part (0-255)

    uint8_t r, g, b;
    switch (segment) {
        case 0: r = 255; b = 0; g = x;       break;  // Red → Yellow
        case 1: g = 255; b = 0; r = 255 - x; break;  // Yellow → Green
        case 2: g = 255; r = 0; b = x;       break;  // Green → Cyan
        case 3: b = 255; r = 0; g = 255 - x; break;  // Cyan → Blue
        case 4: b = 255; g = 0; r = x;       break;  // Blue → Magenta
        case 5: r = 255; g = 0; b = 255 - x; break;  // Magenta → Red
        default:  r = 0; g = 0; b = 0;       break;  // Fallback
    }

    // Pack into RGB565 (5-6-5 bits)
    return ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
}

int main() {
    stdio_init_all();

    gpio_init(pin_d_rsn);
    gpio_set_dir(pin_d_rsn, true);
    gpio_put(pin_d_rsn, 0);
    sleep_ms(10);
    gpio_put(pin_d_rsn, 1);
    sleep_ms(120);

    PIO pio = pio0;
    uint sm = 0;

    uint off_spi = pio_add_program(pio, &spigen_program);
    printf("program spigen loaded at %d\n", off_spi);

    ili9341_program_init(pio, sm, off_spi,
	pin_d_scl, pin_d_sda, pin_d_csn);
    pio_sm_set_enabled(pio, sm, true);


    uint tx_dma_chan = dma_claim_unused_channel(true);
    dma_channel_config dma_cfg = dma_channel_get_default_config(tx_dma_chan);

    channel_config_set_transfer_data_size(&dma_cfg, DMA_SIZE_32);
    channel_config_set_read_increment(&dma_cfg, false); // Repeat same location
    channel_config_set_write_increment(&dma_cfg, false);
    channel_config_set_dreq(&dma_cfg, pio_get_dreq(pio, sm, true));


    uint32_t rddid = ili9341_cmd_rddid(pio, sm, off_spi);
    uint32_t rddst = ili9341_cmd_rddst(pio, sm, off_spi);

    printf("RDDID = %06lX, RDDST = %08lX\n", rddid, rddst);

    // initialize display
    ili9341_cmd(pio, sm, off_spi, ILI9341_SWRESET);
    sleep_ms(150);
    ili9341_cmd(pio, sm, off_spi, ILI9341_SLPOUT);
    sleep_ms(150);

    ili9341_cmd_arg(pio, sm, off_spi, ILI9341_MADCTL, 0x48);
    ili9341_cmd_arg(pio, sm, off_spi, ILI9341_COLMOD, 0x55);

    // config memory window
    ili9341_cmd_caset(pio, sm, off_spi, H_OFFSET, H_OFFSET + DISPLAY_WIDTH - 1);
    ili9341_cmd_raset(pio, sm, off_spi, V_OFFSET, V_OFFSET + DISPLAY_HEIGHT - 1);

    ili9341_cmd(pio, sm, off_spi, ILI9341_DISPON); // Display on
    sleep_ms(50);

    uint32_t frames = 256;

    uint32_t start = time_us_32();
    for (int i=0; i<frames; i++) {
	uint16_t color = hueToRGB565(i*500);
	uint32_t data = (color << 16) | color;
	unsigned size = DISPLAY_WIDTH*DISPLAY_HEIGHT/2;

	// printf("filling with %04X\n",  color);
	spigen_cmd_dma(pio, sm, off_spi,
	    tx_dma_chan, &dma_cfg, &data, size, false);
	sleep_us(10);
    }
    uint32_t end = time_us_32();

    uint32_t delta = end - start;

    printf("%lu frames took %luus, frame time = %.3fms or %ld FPS\n",
	frames, delta, delta * 1e-3 / frames, 1000000 * frames / delta);

    while (true) {
	uint16_t color = hueToRGB565(get_rand_32());
	uint32_t rand = get_rand_32();

	uint16_t x0 = (rand >> 16) % (DISPLAY_WIDTH/2);
	uint16_t x1 = (rand % (DISPLAY_WIDTH/2)) + DISPLAY_WIDTH/2;
	uint16_t y0 = (rand >> 16) % (DISPLAY_HEIGHT/2);
	uint16_t y1 = (rand % (DISPLAY_HEIGHT/2)) + DISPLAY_HEIGHT/2;

	if (((x1-x0) % 2) != 0) {
	    x0 &= ~1; x1 &= ~1;
	}

	uint32_t data = (color << 16) | color;
	unsigned size = (x1-x0)*(y1-y0);

	ili9341_cmd_caset(pio, sm, off_spi, x0, x1);
	ili9341_cmd_raset(pio, sm, off_spi, y0, y1);

	// printf("filling rect %d,%d,%d,%d with %04X\n",
	//     x0, y0, x1, y1, color);
	spigen_cmd_dma(pio, sm, off_spi,
	    tx_dma_chan, &dma_cfg, &data, size, false);
	sleep_us(10);
    }
}

